"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.walk = exports.walkAsync = void 0;
const defaults_1 = require("./defaults");
const helpers_1 = require("./helpers");
const utils_1 = require("./utils");
const node_1 = require("./node");
function matchCallbacks(node, position, ctx) {
    if (!ctx.config.runCallbacks)
        return [];
    if (node.isRoot && !ctx.config.rootObjectCallbacks)
        return [];
    const callbacks = ctx.callbacksByPosition[position];
    if (typeof callbacks == 'undefined') {
        return [];
    }
    return callbacks
        .map(cb => cb)
        .filter(cb => { var _a, _b; return ((_b = (_a = cb.nodeTypeFilters) === null || _a === void 0 ? void 0 : _a.indexOf(node.nodeType)) !== null && _b !== void 0 ? _b : 1) !== -1; })
        .filter(cb => typeof cb.keyFilters === 'undefined'
        || (typeof node.keyInParent !== 'number' && cb.keyFilters.indexOf(node.keyInParent) !== -1));
}
function validateVisitation(node, ctx) {
    const seen = ctx.seenObjects.indexOf(node.val) !== -1;
    if (seen && ctx.config.graphMode === 'graph')
        return false;
    if (seen && ctx.config.graphMode === 'finiteTree')
        throw "The object violates the defined structure. Override 'graphMode' in the config to allow parsing different object structures.";
    ctx.seenObjects.push(node.val);
    return true;
}
function execCallbacks(callbacks, node) {
    callbacks.forEach(cb => {
        cb.callback(node);
        node.executedCallbacks.push(cb);
    });
}
function execCallbacksAsync(callbacks, node) {
    return __awaiter(this, void 0, void 0, function* () {
        for (let i = 0; i < callbacks.length; i++) {
            const cb = callbacks[i];
            yield cb.callback(node);
            node.executedCallbacks.push(cb);
        }
    });
}
function execCallbacksAsyncInParallel(callbacks, node) {
    return __awaiter(this, void 0, void 0, function* () {
        const promises = [];
        for (let i = 0; i < callbacks.length; i++) {
            const cb = callbacks[i];
            const promise = Promise
                .resolve(cb.callback(node))
                .then(() => node.executedCallbacks.push(cb));
            promises.push(promise);
        }
        yield Promise.all(promises);
    });
}
function process(node, mode, ctx) {
    if ((node.nodeType !== 'value') && !validateVisitation(node, ctx))
        return [];
    execCallbacks(matchCallbacks(node, 'preWalk', ctx), node);
    const queue = [];
    const traverse = mode === 'breadth'
        ? (child) => queue.push(child)
        : (child) => process(child, 'depth', ctx);
    node.children.forEach(traverse);
    execCallbacks(matchCallbacks(node, 'postWalk', ctx), node);
    return queue;
}
function processAsync(node, mode, ctx) {
    return __awaiter(this, void 0, void 0, function* () {
        if ((node.nodeType !== 'value') && !validateVisitation(node, ctx))
            return [];
        const asyncExecutor = ctx.config.parallelizeAsyncCallbacks
            ? execCallbacksAsyncInParallel
            : execCallbacksAsync;
        yield asyncExecutor(matchCallbacks(node, 'preWalk', ctx), node);
        const queue = [];
        const traverse = mode === 'breadth'
            ? (child) => __awaiter(this, void 0, void 0, function* () { return Promise.resolve(queue.push(child)); })
            : (child) => __awaiter(this, void 0, void 0, function* () { return yield processAsync(child, 'depth', ctx); });
        const children = node.children;
        for (let i = 0; i < children.length; i++)
            yield traverse(children[i]);
        yield asyncExecutor(matchCallbacks(node, 'postWalk', ctx), node);
        return queue;
    });
}
function buildContext(config) {
    const ctx = defaults_1.buildDefaultContext(config);
    ctx.config.callbacks.forEach(cb => {
        const callback = Object.assign(Object.assign({}, cb), { executionOrder: typeof cb.executionOrder == 'undefined' ? 0 : cb.executionOrder, positionFilters: typeof cb.positionFilters == 'undefined' || cb.positionFilters.length < 1 ? [defaults_1.defaultCallbackPosition] :
                cb.positionFilters });
        callback.positionFilters.forEach(position => {
            if (!(position in ctx.callbacksByPosition))
                ctx.callbacksByPosition[position] = [];
            ctx.callbacksByPosition[position].push(callback);
        });
    });
    for (const key in ctx.callbacksByPosition) {
        ctx.callbacksByPosition[key] = ctx.callbacksByPosition[key].sort(helpers_1.executionOrderSort);
    }
    return ctx;
}
const walkAsync = (obj, config) => __awaiter(void 0, void 0, void 0, function* () {
    const context = buildContext(config);
    const rootNode = node_1.WalkNode.fromRoot(obj);
    try {
        if (context.config.traversalMode === 'depth') {
            yield processAsync(rootNode, 'depth', context);
        }
        else {
            let queue = [rootNode];
            do
                queue = queue.concat(yield processAsync(queue.shift(), 'breadth', context));
            while (queue.length > 0);
        }
    }
    catch (err) {
        if (!(err instanceof utils_1.Break))
            throw err;
    }
    return context;
});
exports.walkAsync = walkAsync;
const walk = (obj, config) => {
    const context = buildContext(config);
    const rootNode = node_1.WalkNode.fromRoot(obj);
    try {
        if (context.config.traversalMode === 'depth') {
            process(rootNode, 'depth', context);
        }
        else {
            let queue = [rootNode];
            do
                queue = queue.concat(process(queue.shift(), 'breadth', context));
            while (queue.length > 0);
        }
    }
    catch (err) {
        if (!(err instanceof utils_1.Break))
            throw err;
    }
    return context;
};
exports.walk = walk;
